﻿:namespace SaltInSpice ⍝ V1.23
⍝ Implement SALT commands within Spice
⍝ 2015 11 12 Adam: moved help layout to framework
⍝ 2015 12 30 Adam: added Test
⍝ 2016 04 07 DanB: Settings example to use ∘
⍝ 2016 11 16 Adam: removed mention of -vars
⍝ 2017 03 01 Adam: updated help to OS appropriate separators
⍝ 2017 07 24 Adam: typo fix
⍝ 2017 09 19 Adam: help text ref to SALT User Guide
⍝ 2018 04 18 Adam: ]??cmd → ]cmd -??
⍝ 2018 04 20 Adam: ]settings settingname -?, add ]Set as alias to ]Settings
⍝ 2018 05 03 Adam: Help tweaks
⍝ 2018 05 31 Adam: [14943] help for track

    ⎕ml← ⎕io←1
    Cnames←'Compare' 'RemoveVersions' 'List' 'Load' 'Save' 'Settings' 'Snap'  'Clean' 'Set'

    WIN←'Win'≡3↑⊃#⎕WG'APLVersion'

    ∇ r←List
      r←⎕NS¨(⍴Cnames)⍴⊂''
      r.(Group Parse)←⊂'SALT' ''
      r.Name←Cnames
      r[1].Desc←'Compare versions of 2 files'
      r[2].Desc←'Remove some or all versions'
      r[3].Desc←'List SALT files or folders in path'
      r[4].Desc←'Load an item (class or namespace or fn or var)'
      r[5].Desc←'Save an item (class or namespace or fn or var)'
      r[6].Desc←'Change local and global settings'
      r[7].Desc←'Save changes made to the workspace'
      r[8].Desc←'Remove all traces of SALT tags in the ws or in specific items'
      r[9].Desc←'Change local and global settings'
    ∇

    ∇ Û←Run(Ûcmd Ûargs);Ûriu
     ⍝ Simulate calling directly from the original ns
      Ûriu←{6::0 ⋄ ##.RIU}0
     ⍝ We can't use :With because if a space is returned
     ⍝ as result it will be lost in :endwith
      ⎕CS ##.THIS  ⍝ We know THIS has been set for us
      Û←⍎'⎕SE.SALT.',Ûcmd,' Ûargs' ⍝ Û vars for Snap
      :If Ûriu<Ûcmd≡'RemoveVersions'
          Û←0 0⍴''
      :EndIf
      :If 'Settings' 'Set'∊⍨⊂Ûcmd
      :AndIf ~Ûriu
          Û←⍕Û
      :EndIf
     
    ∇

    ∇ r←level Help Cmd;setting;settings
      (1↑Cmd)←⎕SE.Dyalog.Utils.ucase 1↑Cmd
      :If 'Clean'≡Cmd
          r←⊂'Removes all traces of SALT tags in the ws or in specific items with the effect that SALT does automatically save changes to any associated file.'
          :If 0=level
              r,←⊂'Args: [items [-deletefiles]]'
          :Else
              r,←'' 'Argument is specific items (default all) in the current namespace'
              r,←'' '-deletefiles also removes the files associated with the items'' use'
              r,←'' 'Examples:' '    Remove all tags in all items in the ws:'('       ]',Cmd)
              r,←'    Remove tags in items A and B and delete their files:'('       ]',Cmd,' A B -deletefiles')
          :EndIf
      :Else
          settings←1 ⎕SE.SALT.Settings'name'
          setting←819⌶##.##.arg∩819⌶⎕A
          :If 'Settings' 'Set'∊⍨⊂Cmd
          :AndIf settings∊⍨⊂setting
              r←⊂']',Cmd,' ',setting,'  ⍝ determines ',(settings⍳⊂setting)⊃1 ⎕SE.SALT.Settings'desc'
              r,←DescSets setting
              :If 'track'≡setting
                  r,←'' 'Possible values are zero or more of:'
                  r,←⊂'  ∘ compiled  preserves the 400⌶ state of SALTed objects (even though 400⌶ directives are not a visible part of a function''s code). The information is reinstated when an object is loaded into the workspace by SALT.'
                  r,←⊂'  ∘ atinfo    retrieves the function, user and timestamp information (as recorded by the monadic system function ⎕AT) pertaining to the last time that the function was saved. The information is reinstated when a function is loaded into the workspace by SALT. Can only be used for traditional functions and operators. This is the default value.'
                  r,←⊂'  ∘ new       tracks unSALTed objects; when these objects are edited using the default editor, they are saved in the first directory named in the workdir session parameter.'
                  r,←''('If no values are specified (]',Cmd,' ""), SALT will only keep the minimum necessary information required to keep track of an object''s origin file.')
                  r,←'' 'Multiple values can be specified in a space or comma separated list. Additional values can be added to the current set by prefixing them with a comma. Active values can be removed by prefixing them with a tilde (~).'
                  r,←'' 'When returning a list of multiple values, space is used as a separator.' ''
              :EndIf
              level←0
          :Else
              r←{(+/∨\' '≠⌽⍵)↑¨↓⍵}⍎'⎕SE.SALT.',{(5×'SALT.'≡5↑⍵)↓⍵}Cmd,' ''?'''
              :If 0=level
                  r←1 0 1/3↑r      ⍝ SALT returns too much help
                  r[1]↓⍨←1+⍴Cmd    ⍝ Strip Cmd name from syntax info
                  r[1],⍨←⊂'Args: ' ⍝ Add label to syntax info
                  :If 'Settings' 'Set'∊⍨⊂Cmd
                      r,←⊂']',Cmd,' workdir -?  ⍝ for description of the specific parameter "workdir"'
                  :EndIf
              :Else  ⍝ More help (?? case)
                  r↓⍨←1                   ⍝ Strip syntax info
                  r←{⍵,⍨'    '/⍨'-'≡⊃⍵}¨r ⍝ Indent -modifier info
                  r,←'' 'Examples:'
                  :Select Cmd
                  :Case 'Save'
                      r,←⊂'    Save "myobj" in <myfile> (not using a version number):'
                      r,←⊂'        ]',Cmd,' myobj /my/folder/myfile'
                      r,←⊂'    Start using version numbers with "myobj":'
                      r,←⊂'        ]',Cmd,' myobj -version'
                      r,←⊂'    Save "myobj" at a specific version number (here 12):'
                      r,←⊂'        ]',Cmd,' myobj -ver=12'
                      r,←⊂'    Save XYZ to folder ensuring folder is created if non existant:'
                      r,←⊂'        ]',Cmd,' XYZ /new/folder/ -makedir'
                      r,←⊂'    Save namespace NN and convert it to script form in the workspace:'
                      r,←⊂'        ]',Cmd,' NN  relative\path\nn  -convert'
                  :Case 'RemoveVersions'
                      r,←⊂'    Remove version 12 of "myfile":'
                      r,←⊂'        ]',Cmd,' /my/folder/myfile -version=12'
                      r,←⊂'    Remove versions above 12 of "myfile":'
                      r,←⊂'        ]',Cmd,' /my/folder/myfile -version=>12'
                      r,←⊂'    Remove versions 12 to 21 of "myfile":'
                      r,←⊂'        ]',Cmd,' /my/folder/myfile -version=12-21'
                      r,←⊂'    Remove versions starting at 12 of "myfile" except the last which will be renamed version 12:'
                      r,←⊂'        ]',Cmd,' /my/folder/myfile -version=≥12 -collapse'
                      r,←⊂'    Remove all versions of "myfile" except the last one which will be renamed version 1:'
                      r,←⊂'        ]',Cmd,' /my/folder/myfile -all -collapse'
                      r,←⊂'    Remove versionning for "myfile" and keep the last one:'
                      r,←⊂'        ]',Cmd,' /my/folder/myfile -all'
                  :Case 'Snap'
                      r,←⊂'    Save every new item in the ws to path specified; changed items are saved to their own location:'
                      r,←⊂'        ]',Cmd,' \new\project'
                      r,←⊂'    Same thing but make sure the folder exists and do not prompt for confirmation of replacement:'
                      r,←⊂'        ]',Cmd,' \new\project -makedir -noprompt'
                      r,←⊂'    Save only namespaces and classes that start with "GUI" and convert namespaces to script form in the ws:'
                      r,←⊂'        ]',Cmd,' /utils/gui -class=9.1,9.4 -pattern=GUI* -convert'
                      r,←⊂'    Save every new item in a file whose prefix is "UTIL" (eg fn <cat> goes to <UTILcat>):'
                      r,←⊂'        ]',Cmd,'p /basics -fileprefix=UTIL'
                      r,←⊂'    Save every item and create a load fn named <lx> to recreate the ws:'
                      r,←⊂'        ]',Cmd,' /my/wsx -loadfn=lx'
                  :Case 'Load'
                      r,←'' '    Load and define the item in a particular file:'
                      r,←⊂'        ]',Cmd,' \tmp\myitem'
                      r,←'' '    Load all version 3 of files starting with TM, searching recursively:'
                      r,←⊂'        ]',Cmd,' TM* -ver=3'
                      r,←'' '    Bring script S, don''t define it, don''t keep a SALT trace and assign its reference to X:'
                      r,←⊂'        ]X←',Cmd,' \tmp\S -noname -nolink'
                      r,←'' '    Disperse the items in /tmp/ns into #.N:'
                      r,←⊂'        ]',Cmd,' /tmp/ns -disperse -target=#.N'
                  :Case 'Compare'
                      r,←⊂'    Compare class in file C with object in workspace:'
                      r,←⊂'        ]SALT.',Cmd,' /my/files/C -version=ws'
                      r,←⊂'    Compare versions 2 and 7 of file X; use +- as line symbols:'
                      r,←⊂'        ]SALT.',Cmd,' \tmp\X -ver=2 7 -symb=+-'
                      r,←⊂'    Compare file A with file B, use always external pgm BC to compare:'
                      r,←⊂'        ]SALT.',Cmd,' /tmp/A  /myfiles/B -use=/bin/BC -permanent'
                  :Case 'Open'⍝ N.B. this command has been moved to its own file
                      r,←⊂'    Open disk explorer to show the contents of /tmp:'
                      r,←⊂'        ]',Cmd,'  /tmp'
                      r,←⊂'    Open file \my\file with Notepad (Windows):'
                      r,←⊂'        ]',Cmd,'  \my\file -with=notepad'
                      r,←⊂'    Open file X with program <vi>:'
                      r,←⊂'        ]',Cmd,' X -with=vi'
                  :Case 'List'
                      r,←⊂'    List all files in \tmp and all files below:'
                      r,←⊂'        ]',Cmd,' \tmp -recursive'
                      r,←⊂'    List folders of folder Y:'
                      r,←⊂'        ]',Cmd,' \myfiles\Y -folders'
                      r,←⊂'    List versions and type of files in relative folder Z:'
                      r,←⊂'        ]',Cmd,'  Z  -type  -ver'
                      r,←⊂'    List the full (starting from #)  path of files:'
                      r,←⊂'        ]',Cmd,'  -full=2'
                  :CaseList 'Settings' 'Set'
                      r,←⊂'    Find the current value of the "workdir" parameter:'
                      r,←⊂'        ]',Cmd,' workdir'
                      r,←⊂'    Overwrite the paths for User Commands:'
                      r,←⊂'        ]',Cmd,'  cmddir  \my\ucmd1',':;'[1+WIN],'\my\ucmd2'
                      r,←⊂'    Set permanently the comparison program (for cmd Compare) to BestC:'
                      r,←⊂'        ]',Cmd,' comp \programs\Bestc.exe  -permanent'
                      r,←⊂'    Remove the requirement to answer prompts when saving items:'
                      r,←⊂'        ]',Cmd,' edprompt 0'
                      r,←'' 'The following parameters impact user commands:'
                      r,←DescSets''
                  :EndSelect
              :EndIf
          :EndIf
      :EndIf
      :If 0=level
          r,←⊂']SALT.',Cmd,' -??      ⍝ for more info and description of all options'
      :EndIf
      r,←''(']',Cmd,' is a thin cover for the ⎕SE.SALT.',Cmd,' function. See the SALT User Guide for full documentation.')
      :If WIN
          r←'/'⎕R'\\'⊢r
      :Else
          r←'\\'⎕R'/'⊢r
      :EndIf
    ∇

    ∇ r←DescSets parameters;a;h
      a←0⍴⊂0⍴⊂''
      h←⊂'    cmddir specifies the full path to the directory (or list of directories) from which to retrieve user commands'
      a,←⊂h
      h←⊂'    debug specifies the level of debugging to use. Possible values are:'
      h,←⊂'        0:   no debugging and report errors in the environment'
      h,←⊂'        >0:  stop if an error is encountered'
      a,←⊂h
      h←⊂'    edprompt specifies whether a user is prompted for confirmation to overwrite the file when modifying a script. Possible values are:'
      h,←⊂'        0:  the user is never prompted for confirmation'
      h,←⊂'        1:  the user is prompted for confirmation each time a script is modified'
      a,←⊂h
      h←⊂'    newcmd specifies when new user commands become effective in the user interface. Possible values are:'
      h,←⊂'        auto    new commands are detected automatically (default on non-Pi)'
      h,←⊂'        manual  new commands do not become effective until the user command ]UReset is run (default on Raspberry Pi)'
      a,←⊂h
      r←⊃,/a/⍨'cmddir' 'debug' 'edprompt' 'newcmd'(∨/⍷)¨⍨⊆parameters
    ∇

    :Section TEST

    ∇ r←Test dummy;T;file;testvar;Eq;ext;old
      r←⍬ ⋄ T←{0::0 ⋄ ⎕SE.UCMD'SALT.',⍺,' ',⍵}
      Eq←{⊃≡/~∘'/\'¨⍺ ⍵~¨⊂'/\'}
      testvar←⎕D
      old←'Settings'T'varfmt xml'
      r,←'/tmp/testvar.dyalog'Eq'Save'T'testvar /tmp/ -nop'
      ⎕EX'testvar'
      r,←(testvar≡⎕D)∧'testvar'≡'Load'T'/tmp/testvar'
      r,←'testvar'≡2⊃2⌷'List'T'/tmp/testvar'
      r,←2=+/'testvar'⍷'Compare'T'testvar -ver=ws'
      r,←(≢'List'T'/tmp/testvar')='RemoveVersions'T'/tmp/testvar -all -nop'
      r,←'xml'≡'Settings'T'varfmt ',old
      r,←1⊣'Snap'T''⍝ not sure how it works
      r,←∨/'Cleanup done'⍷'Clean'T'' ⍝ delete what snap made
     
    ∇
    :EndSection

:Endnamespace ⍝ SaltInSpice  $Revision: 1472 $
