﻿:Namespace format ⍝ V1.29
⍝ Format text vector for std output
⍝ 2015 10 27 Adam: Renamed class
⍝ 2015 11 12 Adam: moved help layout to framework
⍝ 2017 05 26 Adam: typo
⍝ 2018 04 18 Adam: ]??cmd → ]cmd -??
⍝ 2018 05 01 Adam: add SVN tag, help tweaks

    ⎕IO←1 ⋄ ⎕ML←1
    DESC←' paragraph(s) of text into vector of text vectors (or matrix if called from session)'

    ∇ r←List
      r←⎕NS¨⍬ ⍬
   ⍝ Name, group, short description and parsing rules
      r.Name←'Format' 'Layout'
      r.Group←⊂'Output'
      r.Desc←r.Name,¨⊂DESC
      r.Parse←'1L -left=  -right=  -first= ' '1L'
    ∇

    ∇ r←level Help Cmd;⎕ML;call;rm
      ⎕ML←1
      r←⊂Cmd,DESC
      :Select Cmd
      :Case 'Format'
          :If 0=level
              r,←⊂'Arg: text [-first=indent] [-left=indent] [-right=indent]'
              r,←⊂']',Cmd,' -?? ⍝ for more info'
          :Else ⍝ 1≤level
              r,←'' 'Argument is an expression, the result will be formatted. It may be one of the following:'
              :If level∊1 3
                  r,←⊂'    a simple text vector, a vector of text vectors (paragraphs), or a literal text'
                  :If 3=level
                      r,←'' 'Examples:'
                      r,←⊢⊢⊢'    Display current user''s name a quarter screen indented:'('        ]',Cmd,' ⎕AN -l=.25')
                      rm←⍕⎕PW-30
                      r,←⊢⊢⊂'    To indent the first line of each phrase by 2 characters and have a ',rm,'-character right margin:'
                      r,←⊢⊢⊂'         A←''I wandered lonely as a cloud'''
                      r,←⊢⊢⊂'         B←''That floats on high o''er vales and hills,'''
                      r,←⊢⊢⊂'         C←''When all at once I saw a crowd,'''
                      r,←⊢⊢⊂'         D←''A host, of golden daffodils;'''
                      r,←⊢⊢⊂'         ]Format A B C D -first=2 -right=78'
                      call←⊢'        ]Verse←',Cmd,' "''Mary had a little lamb, His fleece was white as snow.''" -f=4 -r=',rm←⍕⎕PW-30
                      r,←⊢⊢⊂'    To set Verse to a nicely formated song, with the first line indented 4 spaces and a right margin of ',(⍕rm),':'
                      r,←call'        )ed Verse   ⍝ to see it'
     
                  :EndIf
              :Else ⍝ 2=level
                  r,←⊂'    expr                  This will be formatted a single paragraph'
                  r,←⊂'    (expr)(expr)(expr)    Each will be formatted separately as a paragraph'
                  r,←⊂'    "''simple text''"       The string will be formatted as a paragraph'
              :EndIf ⍝ Continue 1≤level
              r,←'' '-first= is the first line''s amount of deviation from the left margin. May be:'
              :If level∊1 3
                  r,←⊂'    any integer or ¯1<fraction<1'
              :Else ⍝ 2=level
                  r,←⊂'    any integer           The width of this many characters'
                  r,←⊂'    ¯1<fraction<1         Fraction of the current screen width (⎕PW)'
                  pic←{82=⎕DR'':'.' ⋄ ': ',⎕UCS 9600+⍵} ⍝ Illustrate next lines if Unicode available
                  r,←⊂'    If positive, the first line will be indented further',pic 7 8 8
                  r,←⊂'    If negative, the body text will be hanging',pic 20 8 8 8
              :EndIf
              r,←'' '-left= -right= is inner and outer margin for the body text. May be:'
              :If level∊1 3
                  r,←⊂'    integer≥0 or 0<fraction<1'
                  :If 3=level
                      r,←'' 'Examples:'
                      r,←⊃,/{{(⊂'        ]',⍵),⎕SE.UCMD ⍵}Cmd,⍵}¨' ⎕AN -l=20' ' ⎕AN -l=.5'
                  :EndIf
              :Else ⍝ 2=level
                  r,←⊂'    integer≥0             The width of this many characters'
                  r,←⊂'    0<fraction<1          Fraction of the current screen width (⎕PW)'
              :EndIf
              :If 1≤level
                  r,←⊂''
                  :If 2≠level
                      r,←⊂']',Cmd,' -??? ⍝ for details'
                  :EndIf
                  :If 3≠level
                      r,←⊂']',Cmd,' -???? ⍝ for examples'
                  :EndIf
              :EndIf
          :EndIf
      :Case 'Layout'
          :If 0=level
              r,←⊂'Arg: text'
              r,←⊂']',Cmd,' -?? ⍝ for more info'
          :Else ⍝ 1≤level
              r,←⊂'Argument is both text and layout pattern:'
              r,←⊂'    Leading spaces indicate left margin'
              r,←⊂'    First instance of internal multiple spaces indicates hanging body text'
              r,←⊢'    Trailing spaces indicate right margin' ''
              r,←⊂'Example:'
              r,←⊂'        ⎕PW←42'
              r,←⊂'        ]Box on -fns=on -style=max'
              r,←⊂'        ]',Cmd,' "''   Lorem ipsum:  Dolor sit amet, consectetuer adipiscing elit.''"'
              r,←⊂'┌→───────────────────────────────────────┐'
              r,←⊂'↓   Lorem ipsum:  Dolor sit amet,        │'
              r,←⊂'│                 consectetuer adipiscing│'
              r,←⊂'│                 elit.                  │'
              r,←⊂'└────────────────────────────────────────┘'
     
          :EndIf
      :EndSelect
    ∇

    ∇ r←Run(Cmd Input);text;vals
      text←##.THIS⍎⊃Input.Arguments
      :Select Cmd
      :Case 'Format'
          vals←∊0 Input.Switch¨'left' 'right' 'first'
          r←vals ⎕SE.Dyalog.Utils.formatText text
      :Case 'Layout'
          r←⎕SE.Dyalog.Utils.layoutText text
      :EndSelect
      r←↑⍣(~##.RIU)⊢r
    ∇

:EndNamespace ⍝ format  $Revision$
